@@ -61,6 +61,7 @@
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <glib-object.h>
+#include <glib-unix.h>
 
 #include <gcrypt.h>
 
@@ -77,7 +78,6 @@ typedef int socklen_t;
 #define GKD_COMP_PKCS11     "pkcs11"
 #define GKD_COMP_SECRETS    "secrets"
 #define GKD_COMP_SSH        "ssh"
-#define GKD_COMP_GPG        "gpg"
 
 EGG_SECURE_DECLARE (daemon_main);
 
@@ -111,21 +111,20 @@ static gchar* run_components = DEFAULT_COMPONENTS;
 static gboolean pkcs11_started = FALSE;
 static gboolean secrets_started = FALSE;
 static gboolean ssh_started = FALSE;
-static gboolean gpg_started = FALSE;
 static gboolean dbus_started = FALSE;
 
 static gboolean run_foreground = FALSE;
 static gboolean run_daemonized = FALSE;
 static gboolean run_version = FALSE;
 static gboolean run_for_login = FALSE;
+static gboolean perform_unlock = FALSE;
 static gboolean run_for_start = FALSE;
 static gboolean run_for_replace = FALSE;
 static gchar* login_password = NULL;
 static gchar* control_directory = NULL;
 static guint timeout_id = 0;
 static gboolean initialization_completed = FALSE;
-static gboolean sig_thread_valid = FALSE;
-static pthread_t sig_thread;
+static GMainLoop *loop = NULL;
 
 static GOptionEntry option_entries[] = {
 	{ "start", 's', 0, G_OPTION_ARG_NONE, &run_for_start,
@@ -138,6 +137,8 @@ static GOptionEntry option_entries[] = {
 	  "Run as a daemon", NULL },
 	{ "login", 'l', 0, G_OPTION_ARG_NONE, &run_for_login,
 	  "Run by PAM for a user login. Read login password from stdin", NULL },
+	{ "unlock", 0, 0, G_OPTION_ARG_NONE, &perform_unlock,
+	  "Prompt for login keyring password, or read from stdin", NULL },
 	{ "components", 'c', 0, G_OPTION_ARG_STRING, &run_components,
 	  "The optional components to run", DEFAULT_COMPONENTS },
 	{ "control-directory", 'C', 0, G_OPTION_ARG_FILENAME, &control_directory,
@@ -184,6 +185,14 @@ parse_arguments (int *argc, char** argv[])
 		run_for_start = FALSE;
 	}
 
+	if (run_for_start && perform_unlock) {
+		g_printerr ("gnome-keyring-daemon: The --start option is incompatible with --unlock");
+		perform_unlock = FALSE;
+	}
+
+	if (run_for_login)
+		perform_unlock = TRUE;
+
 	g_option_context_free (context);
 }
 
@@ -285,7 +294,7 @@ log_handler (const gchar *log_domain, GLogLevelFlags log_level,
 		level = LOG_INFO;
 		break;
 	case G_LOG_LEVEL_DEBUG:
-		level = LOG_DEBUG;
+		level = -1;
 		break;
 	default:
 		level = LOG_ERR;
@@ -293,10 +302,12 @@ log_handler (const gchar *log_domain, GLogLevelFlags log_level,
 	}
 
 	/* Log to syslog first */
-	if (log_domain)
-		syslog (level, "%s: %s", log_domain, message);
-	else
-		syslog (level, "%s", message);
+	if (level != -1) {
+		if (log_domain)
+			syslog (level, "%s: %s", log_domain, message);
+		else
+			syslog (level, "%s", message);
+	}
 
 	/* And then to default handler for aborting and stuff like that */
 	g_log_default_handler (log_domain, log_level, message, user_data);
@@ -383,96 +394,29 @@ dump_diagnostics (void)
  * SIGNALS
  */
 
-static sigset_t signal_set;
-static gint signal_quitting = 0;
-
-static gpointer
-signal_thread (gpointer user_data)
-{
-	GMainLoop *loop = user_data;
-	int sig, err;
-
-	for (;;) {
-		err = sigwait (&signal_set, &sig);
-		if (err != EINTR && err != 0) {
-			g_warning ("couldn't wait for signals: %s", g_strerror (err));
-			return NULL;
-		}
-
-		switch (sig) {
-		case SIGUSR1:
-#ifdef WITH_DEBUG
-			dump_diagnostics ();
-#endif /* WITH_DEBUG */
-			break;
-		case SIGPIPE:
-			/* Ignore */
-			break;
-		case SIGHUP:
-		case SIGTERM:
-			g_atomic_int_set (&signal_quitting, 1);
-			g_main_loop_quit (loop);
-			return NULL;
-		default:
-			g_warning ("received unexpected signal when waiting for signals: %d", sig);
-			break;
-		}
-	}
-
-	g_assert_not_reached ();
-	return NULL;
-}
-
-static void
-setup_signal_handling (GMainLoop *loop)
-{
-	int res;
-
-	/*
-	 * Block these signals for this thread, and any threads
-	 * started up after this point (so essentially all threads).
-	 *
-	 * We also start a signal handling thread which uses signal_set
-	 * to catch the various signals we're interested in.
-	 */
-
-	sigemptyset (&signal_set);
-	sigaddset (&signal_set, SIGPIPE);
-	sigaddset (&signal_set, SIGHUP);
-	sigaddset (&signal_set, SIGTERM);
-	sigaddset (&signal_set, SIGUSR1);
-	pthread_sigmask (SIG_BLOCK, &signal_set, NULL);
-
-	res = pthread_create (&sig_thread, NULL, signal_thread, loop);
-	if (res == 0) {
-		sig_thread_valid = TRUE;
-	} else {
-		g_warning ("couldn't startup thread for signal handling: %s",
-		           g_strerror (res));
-	}
-}
-
 void
 gkd_main_quit (void)
 {
-	/*
-	 * Send a signal to terminate our signal thread,
-	 * which in turn runs stops the main loop and that
-	 * starts the shutdown process.
-	 */
-
-	if (sig_thread_valid)
-		pthread_kill (sig_thread, SIGTERM);
-	else
-		raise (SIGTERM);
+	/* Always stop accepting control connections immediately */
+	gkd_control_stop ();
+	g_main_loop_quit (loop);
 }
 
-static void
-cleanup_signal_handling (void)
+static gboolean
+on_signal_term (gpointer user_data)
 {
-	/* The thread is not joinable, so cleans itself up */
-	if (!g_atomic_int_get (&signal_quitting))
-		g_warning ("gkr_daemon_quit() was not used to shutdown the daemon");
+	gkd_main_quit ();
+	g_debug ("received signal, terminating");
+	return FALSE;
+}
+
+static gboolean
+on_signal_usr1 (gpointer user_data)
+{
+#ifdef WITH_DEBUG
+	dump_diagnostics ();
+#endif
+	return TRUE;
 }
 
 /* -----------------------------------------------------------------------------
@@ -549,13 +493,12 @@ clear_login_password (void)
 }
 
 static void
-print_environment (pid_t pid)
+print_environment (void)
 {
 	const gchar **env;
 	for (env = gkd_util_get_environment (); *env; ++env)
 		printf ("%s\n", *env);
-	if (pid)
-		printf ("GNOME_KEYRING_PID=%d\n", (gint)pid);
+	fflush (stdout);
 }
 
 static gboolean
@@ -588,7 +531,7 @@ replace_daemon_at (const gchar *directory)
 
 	/*
 	 * The first control_directory is the environment one, always
-	 * prefer that since it's the one that ssh and gpg will connect to
+	 * prefer that since it's the one that ssh will connect to
 	 */
 	if (control_directory == NULL)
 		control_directory = g_strdup (directory);
@@ -632,6 +575,16 @@ discover_other_daemon (DiscoverFunc callback, gboolean acquire)
 			return TRUE;
 	}
 
+	/* Or the default location when no evironment variable */
+	control_env = g_getenv ("XDG_RUNTIME_DIR");
+	if (control_env) {
+		control = g_build_filename (control_env, "keyring", NULL);
+		ret = (callback) (control);
+		g_free (control);
+		if (ret == TRUE)
+			return TRUE;
+	}
+
 	/* See if we can contact a daemon running, that didn't set an env variable */
 	if (acquire && !gkd_dbus_singleton_acquire (&acquired))
 		return FALSE;
@@ -659,7 +612,7 @@ fork_and_print_environment (void)
 	int fd, i;
 
 	if (run_foreground) {
-		print_environment (getpid ());
+		print_environment ();
 		return;
 	}
 
@@ -680,8 +633,8 @@ fork_and_print_environment (void)
 				exit (WEXITSTATUS (status));
 
 		} else {
-			/* Not double forking, we know the PID */
-			print_environment (pid);
+			/* Not double forking */
+			print_environment ();
 		}
 
 		/* The initial process exits successfully */
@@ -711,8 +664,8 @@ fork_and_print_environment (void)
 			if (pid == -1)
 				exit (1);
 
-			/* We've done two forks. Now we know the PID */
-			print_environment (pid);
+			/* We've done two forks. */
+			print_environment ();
 
 			/* The intermediate child exits */
 			exit (0);
@@ -754,20 +707,6 @@ gkr_daemon_startup_steps (const gchar *components)
 	}
 #endif
 
-#ifdef WITH_GPG
-	if (strstr (components, GKD_COMP_GPG)) {
-		if (gpg_started) {
-			g_message ("The GPG agent was already initialized");
-		} else {
-			gpg_started = TRUE;
-			if (!gkd_daemon_startup_gpg ()) {
-				gpg_started = FALSE;
-				return FALSE;
-			}
-		}
-	}
-#endif
-
 	return TRUE;
 }
 
@@ -868,18 +807,9 @@ on_login_timeout (gpointer data)
 	return FALSE;
 }
 
-static gboolean
-on_idle_initialize (gpointer data)
-{
-	gkr_daemon_initialize_steps (run_components);
-	return FALSE; /* don't run again */
-}
-
 int
 main (int argc, char *argv[])
 {
-	GMainLoop *loop;
-
 	/*
 	 * The gnome-keyring startup is not as simple as I wish it could be.
 	 *
@@ -953,7 +883,7 @@ main (int argc, char *argv[])
 			 * Another daemon was initialized, print out environment
 			 * for any callers, and quit or go comatose.
 			 */
-			print_environment (0);
+			print_environment ();
 			if (run_foreground)
 				while (sleep(0x08000000) == 0);
 			cleanup_and_exit (0);
@@ -978,10 +908,13 @@ main (int argc, char *argv[])
 	if (!gkd_control_listen ())
 		return FALSE;
 
-	/* The --login option. Delayed initialization */
-	if (run_for_login) {
+	if (perform_unlock) {
 		login_password = read_login_password (STDIN);
 		atexit (clear_login_password);
+	}
+
+	/* The --login option. Delayed initialization */
+	if (run_for_login) {
 		timeout_id = g_timeout_add_seconds (LOGIN_TIMEOUT, (GSourceFunc) on_login_timeout, NULL);
 
 	/* Not a login daemon. Startup stuff now.*/
@@ -991,27 +924,43 @@ main (int argc, char *argv[])
 			cleanup_and_exit (1);
 	}
 
+	signal (SIGPIPE, SIG_IGN);
+
 	/* The whole forking and daemonizing dance starts here. */
 	fork_and_print_environment();
 
-	setup_signal_handling (loop);
+	g_unix_signal_add (SIGTERM, on_signal_term, loop);
+	g_unix_signal_add (SIGHUP, on_signal_term, loop);
+	g_unix_signal_add (SIGUSR1, on_signal_usr1, loop);
 
 	/* Prepare logging a second time, since we may be in a different process */
 	prepare_logging();
 
 	/* Remainder initialization after forking, if initialization not delayed */
-	if (!run_for_login)
-		g_idle_add (on_idle_initialize, NULL);
+	if (!run_for_login) {
+		gkr_daemon_initialize_steps (run_components);
+
+		/*
+		 * Close stdout and so that the caller knows that we're
+		 * all initialized, (when run in foreground mode).
+		 *
+		 * However since some logging goes to stdout, redirect that
+		 * to stderr. We don't want the caller confusing that with
+		 * valid output anyway.
+		 */
+		if (dup2 (2, 1) < 1)
+			g_warning ("couldn't redirect stdout to stderr");
+
+		g_debug ("initialization complete");
+	}
 
 	g_main_loop_run (loop);
 
 	/* This wraps everything up in order */
 	egg_cleanup_perform ();
 
-	/* Wrap up signal handling here */
-	cleanup_signal_handling ();
-
 	g_free (control_directory);
 
+	g_debug ("exiting cleanly");
 	return 0;
 }
